﻿// Main firework object in the center
FW_mainFirework = null;

FW_sin_table = [];
FW_cos_table = [];
FW_cos_abs_cube_root_table = [];

FW_sin = function(deg) {
  return FW_sin_table[Math.floor(deg)];
};

FW_cos = function(deg) {
  return FW_cos_table[Math.floor(deg)];
};

FW_cos_abs_cube_root = function(deg) {
  return FW_cos_abs_cube_root_table[Math.floor(deg)];
};

FW_initializeLookups = function() {
  for (var i = -180; i <= 180; ++i) {
    var rad =  i * Math.PI / 180;
    FW_sin_table[i] = Math.sin(rad);
    FW_cos_table[i] = Math.cos(rad);
    FW_cos_abs_cube_root_table[i] = Math.pow(Math.abs(FW_cos_table[i]), 5/6);
  }
};

/**
 * Called by view.onopen
 */
FW_onOpen = function() {
  options.putDefaultValue('initial1', '');
  options.putDefaultValue('initial2', '');
  FW_initializeLookups();
  // Create initial firework objects
  if (options.getValue('notFirstTime')) {
    FW_begin(); 
  } else {
    view.onoptionchanged = FW_onOptionChanged;	  
    sky.appendElement('<a height="16" onclick="plugin.ShowOptionsDialog()" width="120" x="10" y="10" color="#FFFFFF" bold="true">Click here to start</a>'); 
  }
};

FW_begin = function() {
  FW_mainFirework = FW_createFirework();  

  // Setup run-forever timer
  view.setInterval(FW_onStep, 250);
}

FW_onOptionChanged = function() {
  var key = event.propertyName;
  if (key == 'notFirstTime') {
    view.onoptionchanged = '';
    FW_begin();
  }
};


/**
 * Called by the timer to create Firework objects when neeeded and 
 * move all Firework objects to the next animation frame
 */
FW_onStep = function() {
  // If fireworks indicate they are finished
  // Create and assign brand new firework objects
  if (FW_mainFirework.done) {
    FW_mainFirework = FW_createFirework();  
  }

  // Clear all display objects
  FW_clear();
    
  // Call next step in animation
  FW_mainFirework.step();
};

/**
 * Clears all elements in the container div
 */
FW_clear = function() {
  // NOTE: This is the preferred way to enumerate children of an element
  var e = new Enumerator(sky.children);
  
  while (!e.atEnd()) {
    sky.removeElement(e.item());
    e.moveNext();
  }
};

FW_getText = function() {
  var initial1 = options.getValue('initial1');
  var initial2 = options.getValue('initial2');
  
  if (initial1 && initial2) {
    return initial1 + '\n+\n' + initial2;
  }
  return '';
};

/**
 * A fireworks factory!
 * 
 * @return {Object} A CompoundFirework object with randomly configured 
 * fireworks
 */
FW_createFirework = function() {
  // Available colors
  var colors = ['#FF0000',
                '#FF69B4',
		'#FF99FF',
                '#FFFFFF'];                
  
  // Available radii (in pixels)
  var sizes = [80];
  
  // How many particles to explode to
  var particleCounts = [32];                                        

  var x = view.width / 2;
  var beginY = view.height;
  var size = sizes[Utils.getRandom(0, sizes.length)];
  var particleCount = particleCounts[Utils.getRandom(0, particleCounts.length)];    
  var explodeY = view.height - 80;
  var steps = Utils.getRandom(32, 37);
  
  var numFireworks = Utils.getRandom(1, 2);
  var fireworks = [];
  
  for (var i = 0; i < numFireworks; ++i) {
    var color = colors[Utils.getRandom(0, colors.length)];
    fireworks.push(new Firework(x, 
                                beginY, 
                                color,
                                size - (i*5),
                                particleCount,
                                explodeY,
                                steps,
				FW_getText()));
  }
  
  return new CompoundFirework(fireworks);  
};

/**
 * @constructor CompoundFirework
 * @param {Array} fireworks an Array of Firework objects
 */
function CompoundFirework(fireworks) {
  this.fireworks = fireworks;
  this.done = false;
}

/**
 * Animate the next step in the animation
 */
CompoundFirework.prototype.step = function() {
  if (this.done) {
    return;
  }
  
  var fireworksDone = true;
  
  for (var i = 0; i < this.fireworks.length; ++i) {
    var firework = this.fireworks[i];
    
    if (!firework.done) {
      firework.step();
      fireworksDone = false;
    }
  }
  
  if (fireworksDone) {
    this.done = true;
  }  
};


/**
 * @constructor Firework
 * @param {Number} x Initial x coor
 * @param {Number} y Initial y coor
 * @param {String} color Color (6 byte hex string, e.g. #FFFFFF)
 * @param {Number} radius Explosion radius in pixels
 * @param {Number} pieces Number of particles to explode to
 * @param {Number} explodeY Y coor to explode at
 * @param {Number} explodeSteps Number of steps in explosion animation
 */
function Firework(x, y, color, radius, pieces, explodeY, explodeSteps, text) {
  Firework.SHOOTING = 0;
  Firework.SHOOTING_SIZE = 2;
  Firework.SHOOTING_STEPS = 8;

  Firework.EXPLODING = 1;
  Firework.EXPLODING_SIZE = 2;
  
  this.x = x;
  this.y = y;  
  this.color = color;
  this.radius = radius;
  this.pieces = pieces;
  this.explodeY = explodeY;
  this.explodeSteps = explodeSteps;
  this.text = text;
  
  this.currentStep = 0;
  this.particles = [];
  this.done = false;
  this.state = Firework.SHOOTING;
  
  this.particles.push(new Particle(this.color, 
                                   Firework.SHOOTING_SIZE,
                                   this.x, 
                                   this.y, 
                                   this.x, 
                                   this.explodeY, 
                                   Firework.SHOOTING_STEPS));
}

/**
 * Animate the next step in the animation
 */
Firework.prototype.step = function() {
  if (this.done) {
    return;
  }
  
  var particlesDone = true;
  
  for (var i = 0; i < this.particles.length; ++i) {
    var particle = this.particles[i];
    
    if (!particle.done) {
      particle.step();
      particle.draw();
      particlesDone = false;
    }
  }
  
  if (particlesDone) {
    if (this.state == Firework.SHOOTING) {
      this.state = Firework.EXPLODING;
      this.explode();
    } else {
      this.done = true;
    }
  }
};

/**
 * Begin the explosion animation
 */
Firework.prototype.explode = function() {
  var tStep = 360 / this.pieces;
  var t = -180;
  
  if (this.text) {
    this.particles.push(new Text(this.color, this.text, this.x, this.explodeY-2, 50, 60, 1, 13, this.explodeSteps, true));
  }
  
  for (var i = 0; i <= this.pieces; ++i) {
    var x = this.radius * FW_cos(t);
    var y = -(this.radius * .70) * (FW_sin(t) + FW_cos_abs_cube_root(t));
    x += this.x;
    y += this.explodeY;
    
    this.particles.push(new Particle(this.color, 
                                     Firework.EXPLODING_SIZE, 
                                     this.x, 
                                     this.explodeY, 
                                     x, 
                                     y, 
                                     this.explodeSteps, 
                                     true));   
    
    t += tStep;
  }  
};


/**
 * @constructor Particle
 * @param {String} color Color (6 byte hex string, e.g. #FFFFFF)
 * @param {Number} size Width/height of the particle
 * @param {Number} x Initial x coor
 * @param {Number} y Initial y coor 
 * @param {Number} toX Initial Final x coor
 
 * @param {Number} toY Initial Final y coor 
 * @param {Number} length Number of steps in the animation
 * @param {Boolean} explodeSteps Fade out while animating?
 */
function Particle(color, size, x, y, toX, toY, length, fadeOut) {
  this.color = color;
  this.size = size;
  this.x = x;
  this.y = y;
  this.toX = toX;
  this.toY = toY;
  this.length = length;
  
  this.fadeOut = fadeOut;
  
  this.xDelta = (this.toX - this.x) / this.length;
  this.yDelta = (this.toY - this.y) / this.length;
  
  this.done = false;
  this.currentStep = 0;
}

/**
 * Animate the next step in the animation
 */
Particle.prototype.step = function() {
  if (this.done) {
    return;
  }
  
  this.x += this.xDelta;
  this.y += this.yDelta;  
  this.draw();
  
  ++this.currentStep;
  
  if (this.currentStep >= this.length) {
    this.done = true;
  }
};

/**
 * Draw the particle
 */
Particle.prototype.draw = function() {
  sky.appendElement(this.createElement());
};

/**
 * Create the XML string for the element to be added to the sky
 * @return {String} The XML string
 */
Particle.prototype.createElement = function() {
  var s = '<div ';
  s += 'x="' + this.x + '" ';
  s += 'y="' + this.y + '" ';
  s += 'background="' + this.color + '" ';
  s += 'width="' + this.size + '" ';
  s += 'height="' + this.size + '" ';
  if (this.fadeOut) {
    s += 'opacity="' + (255 - 255*(this.currentStep/this.length)) + '" ';  
  }
  s += '/>';
  return s;
};


function Text(color, text, x, y, width, height, initialSize, finalSize, length, fadeOut) {
  this.color = color;
  this.text = text;
  this.width = width;
  this.height = height;
  this.x = x - (width / 2);
  this.y = y - (height / 2);
  this.initialSize = initialSize;
  this.finalSize = finalSize;
  this.length = length;  
  this.fadeOut = fadeOut;
  
  this.size = initialSize;
  this.sizeDelta = (this.finalSize - this.initialSize) / this.length;
  
  this.done = false;
  this.currentStep = 0;
}

/**
 * Animate the next step in the animation
 */
Text.prototype.step = function() {
  if (this.done) {
    return;
  }
  
  this.size += this.sizeDelta;

  if ((this.currentStep / this.length) > 0.5) {
    this.draw();
  }  
  
  ++this.currentStep;
  
  if (this.currentStep >= this.length) {
    this.done = true;
  }
};

/**
 * Draw the particle
 */
Text.prototype.draw = function() {
  sky.appendElement(this.createElement());
};

/**
 * Create the XML string for the element to be added to the sky
 * @return {String} The XML string
 */
Text.prototype.createElement = function() {
  var s = '<label ';
  s += 'x="' + this.x + '" ';
  s += 'y="' + this.y + '" ';
  s += 'color="' + this.color + '" ';
  s += 'width="' + this.width + '" ';
  s += 'height="' + this.height + '" ';
  s += 'align="center" ';
  s += 'vAlign="middle" ';
  s += 'bold="true" ';  
  s += 'size="' + this.size + '" ';
  if (this.fadeOut) {
    s += 'opacity="' + (255 - 200*(this.currentStep/this.length)) + '" ';  
  }
  s += '>';
  s += this.text;
  s += '</label>';
  return s;
};


/**
 * Utils namespace
 */
var Utils = {};

/**
 * Returns a random integer between min (inclusive) and max (exclusive)
 * Max is exclusive because it's more convenient for our purposes
 * @param {Number} min
 * @param {Number} max
 */
Utils.getRandom = function(min, max) {
  var range = max - min;
  
  return min + Math.floor(Math.random() * range);
};
